#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: python/SHE_ArCTIC/ImageIO.py

Created on: 04/23/18
Author: James Nightingale
"""

import os

import numpy as np
from astropy.io import fits


def make_path_if_does_not_exist(path):
    if os.path.exists(path) is False:
        os.makedirs(path)


def numpy_array_to_fits(array, file_path, overwrite=False):
    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    new_hdr = fits.Header()
    hdu = fits.PrimaryHDU(array, new_hdr)
    hdu.writeto(file_path)


def numpy_array_from_fits(file_path, hdu):
    hdu_list = fits.open(file_path)
    return np.array(hdu_list[hdu].data)



import numpy as np

from autocti.charge_injection import ci_data
from autocti.charge_injection.plotters import ci_data_plotters
from autocti.data import mask as msk
from test.charge_injection.plotters.fixtures import *
from test.mock.mock import MockGeometry, MockPattern


@pytest.fixture(name='data_plotter_path')
def make_ci_data_plotter_setup():
    return "{}/../../test_files/plotting/ci_data/".format(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture(name='mask')
def make_mask():
    return msk.Mask.empty_for_shape(shape=(6, 6))


@pytest.fixture(name='image')
def make_image():
    return np.ones((6, 6))


@pytest.fixture(name='noise_map')
def make_noise_map():
    return 2.0 * np.ones((6, 6))


@pytest.fixture(name='ci_pre_cti')
def make_ci_pre_cti():
    return 3.0 * np.ones((6, 6))


@pytest.fixture(name='data')
def make_ci_data(image, noise_map, ci_pre_cti):
    return ci_data.CIData(image=image, noise_map=noise_map, ci_pre_cti=ci_pre_cti, ci_pattern=MockPattern(),
                          ci_frame=MockGeometry())


def test__ci_sub_plot_output(data, data_plotter_path, plot_patch):
    ci_data_plotters.plot_ci_subplot(ci_data=data, extract_array_from_mask=True,
                                     cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                     output_path=data_plotter_path, output_format='png')

    assert data_plotter_path + 'ci_data.png' in plot_patch.paths


def test__ci_individuals__output_dependent_on_inputs(data, data_plotter_path, plot_patch):
    ci_data_plotters.plot_ci_data_individual(ci_data=data, extract_array_from_mask=True,
                                             should_plot_image=True,
                                             should_plot_ci_pre_cti=True,
                                             output_path=data_plotter_path, output_format='png')

    assert data_plotter_path + 'ci_image.png' in plot_patch.paths

    assert data_plotter_path + 'ci_noise_map.png' not in plot_patch.paths

    assert data_plotter_path + 'ci_pre_cti.png' in plot_patch.paths

    assert data_plotter_path + 'ci_signal_to_noise_map.png' not in plot_patch.paths


def test__image_is_output(data, mask, data_plotter_path, plot_patch):
    ci_data_plotters.plot_image(ci_data=data, mask=mask, extract_array_from_mask=True,
                                cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                output_path=data_plotter_path, output_format='png')
    assert data_plotter_path + 'ci_image.png' in plot_patch.paths


def test__noise_map_is_output(data, mask, data_plotter_path, plot_patch):
    ci_data_plotters.plot_noise_map(ci_data=data, mask=mask, extract_array_from_mask=True,
                                    cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                    output_path=data_plotter_path, output_format='png')
    assert data_plotter_path + 'ci_noise_map.png' in plot_patch.paths


def test__ci_pre_cti_is_output(data, mask, data_plotter_path, plot_patch):
    ci_data_plotters.plot_ci_pre_cti(ci_data=data, mask=mask, extract_array_from_mask=True,
                                     cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                     output_path=data_plotter_path, output_format='png')
    assert data_plotter_path + 'ci_pre_cti.png' in plot_patch.paths


def test__signal_to_noise_map_is_output(data, mask, data_plotter_path, plot_patch):
    ci_data_plotters.plot_signal_to_noise_map(ci_data=data, mask=mask, extract_array_from_mask=True,
                                              cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                              output_path=data_plotter_path, output_format='png')
    assert data_plotter_path + 'ci_signal_to_noise_map.png' in plot_patch.paths



from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autofit.mapper import prior_model
from autocti.charge_injection import ci_hyper
from autocti.pipeline import pipeline as pl
from autocti.pipeline import phase as ph
from autocti.model import arctic_params

# In this pipeline, we'll perform an analysis which fits three parallel trap species to a set of charge
# injection imaging data. This will include a hyper-phase which scales the noise in the analysis, to prevent
# over-fitting the highest S/N charge injection images. The pipeline uses four phases:

# Phase 1) Fit a small section (60 columns of every charge injection) using a parallel CTI model
#          with 1 trap species and a model for the parallel CCD volume filling parameters.

# Phase 2) Fit a small section (again, 60 columns) using a parallel CTI model with 3 trap species and a model for the
#          parallel CCD volume filling parameters. The priors on trap densities and volume filling parameters are
#          initialized from the results of phase 1.

# Phase 3) Use the best-fit model from phase 2 to scale the noise of each image, to ensure that the higher and
#          lower S/N images are weighted more equally in their contribution to the likelihood.

# Phase 4) Refit the phase 2 model, using priors initialized from the results of phase 2 and the scaled noise-map
#          computed in phase 3.

def make_pipeline(pipeline_path=''):

    pipeline_name = 'pipeline_parallel_x3_species'
    pipeline_path = pipeline_path + pipeline_name

    ### PHASE 1 ###

    # In phase 1, we will fit the data with a one species parallel CTI model and parallel CCD filling model. In this
    # phase we will:

    # 1) Extract and fit the 10 columns of each charge injection region which is furthest from the clocking direction
    # (and therefore least affected by parallel CTI).

    class ParallelPhase(ph.ParallelPhase):

        def pass_priors(self, previous_results):
            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0

    phase1 = ParallelPhase(optimizer_class=nl.MultiNest,
                           parallel_species=[prior_model.PriorModel(arctic_params.Species)],
                           parallel_ccd=arctic_params.CCD, columns=60,
                           phase_name=pipeline_path + '/phase_1_x1_species')

    # You'll see these lines throughout all of the example pipelines. They are used to make MultiNest sample the \
    # non-linear parameter space faster.

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 30
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # In phase 2, we will fit the data with a three species parallel CTI model and parallel CCD filling model. In this
    # phase we will:

    # 1) As in phase 1, extract and fit the 10 columns of charge injection imaging data closest to the read-out
    # register.
    # 2) Use priors on the trap density and ccd volume filling parameters based on the results of phase 1.

    class ParallelPhase(ph.ParallelPhase):

        def pass_priors(self, previous_results):

            previous_total_density = previous_results[-1].constant.parallel_species[0].trap_density

            self.parallel_species[0].trap_density = prior.UniformPrior(lower_limit=0.0, upper_limit=previous_total_density)
            self.parallel_species[1].trap_density = prior.UniformPrior(lower_limit=0.0, upper_limit=previous_total_density)
            self.parallel_species[2].trap_density = prior.UniformPrior(lower_limit=0.0, upper_limit=previous_total_density)
            self.parallel_species[0].trap_lifetime = prior.UniformPrior(lower_limit=0.0, upper_limit=30.0)
            self.parallel_species[1].trap_lifetime = prior.UniformPrior(lower_limit=0.0, upper_limit=30.0)
            self.parallel_species[2].trap_lifetime = prior.UniformPrior(lower_limit=0.0, upper_limit=30.0)

            self.parallel_ccd.well_notch_depth = previous_results[0].variable.parallel_ccd.well_notch_depth
            self.parallel_ccd.well_fill_beta = previous_results[0].variable.parallel_ccd.well_fill_beta
            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0

    phase2 = ParallelPhase(parallel_species=[prior_model.PriorModel(arctic_params.Species),
                                         prior_model.PriorModel(arctic_params.Species),
                                         prior_model.PriorModel(arctic_params.Species)],
                         parallel_ccd=arctic_params.CCD, columns=60, optimizer_class=nl.MultiNest,
                         phase_name=pipeline_path + '/phase_2_x3_species_initialize')

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 30
    phase2.optimizer.sampling_efficiency = 0.2

    ### PHASE 3 ###

    # The best fit model of phase 2 is used to create a 'noise-scaling' map for every charge injection image. These
    # noise-scaling maps are used in conjunction with 'hyper-noise' models to scale the noise-maps in a way that
    # equally weights the fit across all charge injection images.

    class ParallelHyperModelFixedPhase(ph.ParallelHyperPhase):

        def pass_priors(self, previous_results):

            self.parallel_species = previous_results[1].constant.parallel_species
            self.parallel_ccd = previous_results[1].constant.parallel_ccd

    phase3 = ParallelHyperModelFixedPhase(parallel_species=[prior_model.PriorModel(arctic_params.Species),
                                                        prior_model.PriorModel(arctic_params.Species),
                                                        prior_model.PriorModel(arctic_params.Species)],
                                        parallel_ccd=arctic_params.CCD,
                                        hyper_noise_scalar_ci_regions=ci_hyper.CIHyperNoiseScaler,
                                        hyper_noise_scalar_parallel_trails=ci_hyper.CIHyperNoiseScaler,
                                        optimizer_class=nl.MultiNest, columns=None,
                                        phase_name=pipeline_path + '/phase_3_noise_scaling')

    ### PHASE 4 ###

    # In phase 4, we will fit the data with a 3 species parallel CTI model and parallel CCD filling model. In this
    # phase we will:

    # 1) Use the complete charge injection image, as opposed to extracting a sub-set of columns.
    # 2) Use the scaled noise-map computed in phase 3.
    # 3) Initialize the priors on the parallel CTI model from the results of phase 2.

    class ParallelHyperFixedPhase(ph.ParallelHyperPhase):

        def pass_priors(self, previous_results):

            self.hyper_noise_scalar_ci_regions = previous_results[2].constant.hyper_noise_scalar_ci_regions
            self.hyper_noise_scalar_parallel_trails = previous_results[2].constant.hyper_noise_scalar_parallel_trails
            self.parallel_species = previous_results[1].variable.parallel_species
            self.parallel_ccd = previous_results[1].variable.parallel_ccd
            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0

    phase4 = ParallelHyperFixedPhase(parallel_species=[prior_model.PriorModel(arctic_params.Species),
                                                   prior_model.PriorModel(arctic_params.Species),
                                                   prior_model.PriorModel(arctic_params.Species)],
                                   parallel_ccd=arctic_params.CCD,
                                   hyper_noise_scalar_ci_regions=ci_hyper.CIHyperNoiseScaler,
                                   hyper_noise_scalar_parallel_trails=ci_hyper.CIHyperNoiseScaler,
                                   optimizer_class=nl.MultiNest,
                                   phase_name=pipeline_path + '/phase_4_final')

    # For the final CTI model, constant efficiency mode has a tendency to sample parameter space too fast and infer an
    # inaccurate model. Thus, we turn it off for phase 2.

    phase4.optimizer.const_efficiency_mode = False
    phase4.optimizer.n_live_points = 50
    phase4.optimizer.sampling_efficiency = 0.3

    return pl.Pipeline(phase1, phase2, phase3, phase4)


#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: python/VIS_CTI_ChargeInjection/CIImage.pyy

Created on: 02/14/18
Author: James Nightingale
"""

import numpy as np

from autocti.charge_injection import ci_frame as frame
from autocti.charge_injection import ci_pattern as pattern
from autocti.data import cti_image
from autocti.data import mask as msk
from autocti.data import util


class CIData(object):

    def __init__(self, image, noise_map, ci_pre_cti, ci_pattern, ci_frame):
        self.image = image
        self.noise_map = noise_map
        self.ci_pre_cti = ci_pre_cti
        self.ci_pattern = ci_pattern
        self.ci_frame = ci_frame

    @property
    def chinj(self):
        return frame.ChInj(frame_geometry=self.ci_frame, ci_pattern=self.ci_pattern)

    @property
    def shape(self):
        return self.image.shape

    def map_to_ci_data_fit(self, func, mask: msk.Mask):
        """
        Maps an extraction function onto the arrays in this object and a mask.

        Parameters
        ----------
        func
            The extraction function
        mask
            A mask

        Returns
        -------
        masked_ci_data: MaskedCIData
        """
        return MaskedCIData(image=func(self.image),
                            noise_map=func(self.noise_map),
                            ci_pre_cti=func(self.ci_pre_cti),
                            mask=func(mask),
                            ci_pattern=self.ci_pattern,
                            ci_frame=self.ci_frame)

    def map_to_ci_hyper_data_fit(self, func, mask, noise_scaling_maps):
        """
        Maps an extraction function onto the arrays in this object, a mask and noise scaling maps.

        Parameters
        ----------
        func
            The extraction function
        mask
            A mask
        noise_scaling_maps
            A list of noise maps used for scaling noise in poorly fit regions

        Returns
        -------
        masked_ci_data: MaskedCIData
        """
        return MaskedCIHyperData(image=func(self.image),
                                 noise_map=func(self.noise_map),
                                 ci_pre_cti=func(self.ci_pre_cti),
                                 mask=func(mask),
                                 ci_pattern=self.ci_pattern,
                                 ci_frame=self.ci_frame,
                                 noise_scaling_maps=list(map(func, noise_scaling_maps)))

    def parallel_extractor(self, columns):
        """
        Creates a function to extract a parallel section for given columns
        """

        def extractor(obj):
            return self.chinj.parallel_calibration_section_for_columns(array=obj, columns=columns)

        return extractor

    def serial_extractor(self, rows):
        """
        Creates a function to extract a serial section for given rows
        """

        def extractor(obj):
            return self.chinj.serial_calibration_section_for_rows(array=obj, rows=rows)

        return extractor

    def parallel_serial_extractor(self):
        """
        Creates a function to extract a parallel and serial calibration section
        """

        def extractor(obj):
            return self.chinj.parallel_serial_calibration_section(obj)

        return extractor

    def parallel_calibration_data(self, columns: (int,), mask: msk.Mask):
        """
        Creates a MaskedCIData object for a parallel section of the CCD

        Parameters
        ----------
        columns
            Columns to be extracted
        mask
            A mask

        Returns
        -------
        MaskedCIData
        """
        return self.map_to_ci_data_fit(self.parallel_extractor(columns), mask)

    def serial_calibration_data(self, rows: (int,), mask: msk.Mask):
        """
        Creates a MaskedCIData object for a serial section of the CCD

        Parameters
        ----------
        rows
            Rows to be extracted
        mask
            A mask
        Returns
        -------
        MaskedCIData
        """
        return self.map_to_ci_data_fit(self.serial_extractor(rows), mask)

    def parallel_serial_calibration_data(self, mask: msk.Mask):
        """
        Creates a MaskedCIData object for a section of the CCD

        Parameters
        ----------
        mask
            A mask
        Returns
        -------
        MaskedCIData
        """
        return self.map_to_ci_data_fit(self.parallel_serial_extractor(), mask)

    def parallel_hyper_calibration_data(self, columns: (int,), mask: msk.Mask, noise_scaling_maps: (np.ndarray,)):
        """
        Creates a MaskedCIData object for a parallel section of the CCD

        Parameters
        ----------
        noise_scaling_maps
            A list of maps that are used to scale noise
        columns
            Columns to be extracted
        mask
            A mask
        Returns
        -------
        MaskedCIHyperData
        """
        return self.map_to_ci_hyper_data_fit(self.parallel_extractor(columns), mask, noise_scaling_maps)

    def serial_hyper_calibration_data(self, rows: (int,), mask: msk.Mask, noise_scaling_maps: (np.ndarray,)):
        """
        Creates a MaskedCIData object for a serial section of the CCD

        Parameters
        ----------
        noise_scaling_maps
            A list of maps that are used to scale noise
        rows
            Rows to be extracted
        mask
            A mask
        Returns
        -------
        MaskedCIHyperData
        """
        return self.map_to_ci_hyper_data_fit(self.serial_extractor(rows), mask, noise_scaling_maps)

    def parallel_serial_hyper_calibration_data(self, mask: msk.Mask, noise_scaling_maps: (np.ndarray,)):
        """
        Creates a MaskedCIData object for a section of the CCD

        Parameters
        ----------
        noise_scaling_maps
            A list of maps that are used to scale noise
        mask
            A mask
        Returns
        -------
        MaskedCIHyperData
        """
        return self.map_to_ci_hyper_data_fit(self.parallel_serial_extractor(), mask, noise_scaling_maps)

    @property
    def signal_to_noise_map(self):
        """The estimated signal-to-noise_maps mappers of the image."""
        signal_to_noise_map = np.divide(self.image, self.noise_map)
        signal_to_noise_map[signal_to_noise_map < 0] = 0
        return signal_to_noise_map

    @property
    def signal_to_noise_max(self):
        """The maximum value of signal-to-noise_maps in an image pixel in the image's signal-to-noise_maps mappers"""
        return np.max(self.signal_to_noise_map)


class MaskedCIData(object):

    def __init__(self, image, noise_map, ci_pre_cti, mask, ci_pattern, ci_frame):
        """A fitting image is the collection of data components (e.g. the image, noise-maps, PSF, etc.) which are used \
        to generate and fit it with a model image.

        The fitting image is in 2D and masked, primarily to remove cosmic rays.

        The fitting image also includes a number of attributes which are used to performt the fit, including (y,x) \
        grids of coordinates, convolvers and other utilities.

        Parameters
        ----------
        image : im.Image
            The 2D observed image and other observed quantities (noise-map, PSF, exposure-time map, etc.)
        mask: msk.Mask | None
            The 2D mask that is applied to image data.

        Attributes
        ----------
        image : ScaledSquarePixelArray
            The 2D observed image data (not an instance of im.Image, so does not include the other data attributes,
            which are explicitly made as new attributes of the fitting image).
        noise_map : NoiseMap
            An array describing the RMS standard deviation error in each pixel, preferably in units of electrons per
            second.
        mask: msk.Mask
            The 2D mask that is applied to image data.
        """
        self.image = image
        self.noise_map = noise_map
        self.ci_pre_cti = ci_pre_cti
        self.mask = mask
        self.ci_pattern = ci_pattern
        self.ci_frame = ci_frame

    @property
    def chinj(self):
        return frame.ChInj(frame_geometry=self.ci_frame, ci_pattern=self.ci_pattern)

    @property
    def signal_to_noise_map(self):
        """The estimated signal-to-noise_maps mappers of the image."""
        signal_to_noise_map = np.divide(self.image, self.noise_map)
        signal_to_noise_map[signal_to_noise_map < 0] = 0
        return signal_to_noise_map

    @property
    def signal_to_noise_max(self):
        """The maximum value of signal-to-noise_maps in an image pixel in the image's signal-to-noise_maps mappers"""
        return np.max(self.signal_to_noise_map)


class MaskedCIHyperData(MaskedCIData):
    def __init__(self, image, noise_map, ci_pre_cti, mask, ci_pattern, ci_frame, noise_scaling_maps):
        super().__init__(image, noise_map, ci_pre_cti, mask, ci_pattern, ci_frame)
        self.noise_scaling_maps = noise_scaling_maps


def simulate(ci_pre_cti, frame_geometry, ci_pattern, cti_params, cti_settings, read_noise=None, cosmic_ray_image=None,
             noise_seed=-1):
    """Simulate a charge injection image, including effects like noises.

    Parameters
    -----------
    ci_pre_cti
    cosmic_ray_image
        The dimensions of the output simulated charge injection image.
    frame_geometry : ci_frame.CIQuadGeometry
        The quadrant geometry of the simulated image, defining where the parallel / serial overscans are and \
        therefore the direction of clocking and rotations before input into the cti algorithm.
    ci_pattern : ci_pattern.CIPatternSimulate
        The charge injection ci_pattern (regions, normalization, etc.) of the charge injection image.
    cti_params : ArcticParams.ArcticParams
        The CTI model parameters (trap density, trap lifetimes etc.).
    cti_settings : ArcticSettings.ArcticSettings
        The settings that control the cti clocking algorithm (e.g. ccd well_depth express option).
    read_noise : None or float
        The FWHM of the Gaussian read-noises added to the image.
    noise_seed : int
        Seed for the read-noises added to the image.
    """

    ci_frame = frame.ChInj(frame_geometry=frame_geometry, ci_pattern=ci_pattern)

    if cosmic_ray_image is not None:
        ci_pre_cti += cosmic_ray_image

    ci_pre_cti = cti_image.ImageFrame(frame_geometry=frame_geometry, array=ci_pre_cti)

    ci_post_cti = ci_pre_cti.add_cti_to_image(cti_params, cti_settings)

    if read_noise is not None:
        ci_image = ci_post_cti + read_noise_map_from_shape_and_sigma(shape=ci_post_cti.shape, sigma=read_noise,
                                                                     noise_seed=noise_seed)
        ci_noise_map = read_noise * np.ones(ci_post_cti.shape)
    else:
        ci_image = ci_post_cti
        ci_noise_map = None

    return CIData(ci_frame=ci_frame, image=ci_image, noise_map=ci_noise_map, ci_pre_cti=ci_pre_cti,
                  ci_pattern=ci_pattern)


def ci_pre_cti_from_ci_pattern_geometry_image_and_mask(ci_pattern, image, mask=None):
    """Setup a pre-cti image from this charge injection ci_data, using the charge injection ci_pattern.

    The pre-cti image is computed depending on whether the charge injection ci_pattern is uniform, non-uniform or \
    'fast' (see ChargeInjectPattern).
    """
    if isinstance(ci_pattern, pattern.CIPatternUniform):
        return ci_pattern.ci_pre_cti_from_shape(image.shape)
    return ci_pattern.ci_pre_cti_from_ci_image_and_mask(image, mask)


def load_ci_data_from_fits(frame_geometry, ci_pattern,
                           ci_image_path, ci_image_hdu=0,
                           ci_noise_map_path=None, ci_noise_map_hdu=0,
                           ci_noise_map_from_single_value=None,
                           ci_pre_cti_path=None, ci_pre_cti_hdu=0,
                           mask=None):
    ci_image = util.numpy_array_from_fits(file_path=ci_image_path, hdu=ci_image_hdu)

    if ci_noise_map_path is not None:
        ci_noise_map = util.numpy_array_from_fits(file_path=ci_noise_map_path, hdu=ci_noise_map_hdu)
    else:
        ci_noise_map = np.ones(ci_image.shape) * ci_noise_map_from_single_value

    if ci_pre_cti_path is not None:
        ci_pre_cti = util.numpy_array_from_fits(file_path=ci_pre_cti_path, hdu=ci_pre_cti_hdu)
    else:
        ci_pre_cti = ci_pre_cti_from_ci_pattern_geometry_image_and_mask(ci_pattern, ci_image, mask=mask)

    return CIData(image=ci_image, noise_map=ci_noise_map, ci_pre_cti=ci_pre_cti, ci_pattern=ci_pattern,
                  ci_frame=frame_geometry)


def read_noise_map_from_shape_and_sigma(shape, sigma, noise_seed=-1):
    """Generate a two-dimensional read noises-map, generating values from a Gaussian distribution with mean 0.0.

    Params
    ----------
    shape : (int, int)
        The (x,y) image_shape of the generated Gaussian noises map.
    read_noise : float
        Standard deviation of the 1D Gaussian that each noises value is drawn from
    seed : int
        The seed of the random number generator, used for the random noises maps.
    """
    if noise_seed == -1:
        # Use one seed, so all regions have identical column non-uniformity.
        noise_seed = np.random.randint(0, int(1e9))
    np.random.seed(noise_seed)
    read_noise_map = np.random.normal(loc=0.0, scale=sigma, size=shape)
    return read_noise_map



from matplotlib import pyplot as plt

from autocti.charge_injection.plotters import fit_plotters
from autocti.data.plotters import plotter_util


def plot_fit_subplot(
        fit, extract_array_from_mask=False,
        figsize=None, aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
        output_path=None, output_filename='ci_fit', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    xyticksize
    ylabelsize
    xlabelsize
    titlesize
    cb_tick_labels
    cb_tick_values
    cb_pad
    cb_fraction
    cb_ticksize
    linscale
    linthresh
    norm_max
    norm_min
    norm
    cmap
    aspect
    figsize
    extract_array_from_mask
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=9)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    fit_plotters.plot_image(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 2)

    fit_plotters.plot_noise_map(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 3)

    fit_plotters.plot_signal_to_noise_map(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 4)

    fit_plotters.plot_ci_pre_cti(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 5)

    fit_plotters.plot_ci_post_cti(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 7)

    fit_plotters.plot_residual_map(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plt.subplot(rows, columns, 8)

    fit_plotters.plot_chi_squared_map(
        fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad, cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fit_individuals(
        fit, extract_array_from_mask=False,
        should_plot_image=False,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_ci_pre_cti=False,
        should_plot_ci_post_cti=False,
        should_plot_residual_map=False,
        should_plot_chi_squared_map=False,
        output_path=None, output_format='show'):
    if should_plot_image:
        fit_plotters.plot_image(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_noise_map:
        fit_plotters.plot_noise_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_signal_to_noise_map:
        fit_plotters.plot_signal_to_noise_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_ci_pre_cti:
        fit_plotters.plot_ci_pre_cti(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_ci_post_cti:
        fit_plotters.plot_ci_post_cti(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_residual_map:
        fit_plotters.plot_residual_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)

    if should_plot_chi_squared_map:
        fit_plotters.plot_chi_squared_map(
            fit=fit, mask=fit.mask, extract_array_from_mask=extract_array_from_mask,
            output_path=output_path, output_format=output_format)



import numpy as np
import pytest
from autofit.tools import fit_util

from autocti.charge_injection import ci_fit
from autocti.charge_injection import ci_frame
from autocti.charge_injection import ci_hyper, ci_data
from test.mock.mock import MockGeometry, MockPattern, MockCIFrame, MockCIPreCTI, MockParams, MockSettings


@pytest.fixture(name='ci_data_fit')
def make_ci_data_fit():
    ci_image = 1.0 * np.ones((2, 2))
    ci_noise_map = 2.0 * np.ones((2, 2))

    ci_pre_cti = MockCIPreCTI(frame_geometry=MockGeometry(), ci_pattern=MockPattern(),
                              array=3.0 * np.ones((2, 2)), value=3.0)
    ci_mask = np.ma.zeros((2, 2))

    return ci_data.MaskedCIData(ci_frame=MockCIFrame(value=3.0), ci_pattern=MockPattern(), image=ci_image,
                                noise_map=ci_noise_map, ci_pre_cti=ci_pre_cti, mask=ci_mask)


@pytest.fixture(name='ci_data_hyper_fit')
def make_ci_datas_hyper_fit(ci_data_fit):
    ci_noise_scalings = [np.array([[1.0, 2.0], [3.0, 4.0]]), np.array([[1.0, 2.0], [3.0, 4.0]])]

    return ci_data.MaskedCIHyperData(
        ci_frame=MockCIFrame(value=3.0), ci_pattern=MockPattern(), image=ci_data_fit.image,
        noise_map=ci_data_fit.noise_map, ci_pre_cti=ci_data_fit.ci_pre_cti, mask=ci_data_fit.mask,
        noise_scaling_maps=ci_noise_scalings)


class TestCIFit:
    class TestFits:

        def test__residual_map_same_as_calculated_individually(self, ci_data_fit):
            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=ci_data_fit.image,
                                                                               mask=ci_data_fit.mask,
                                                                               model_data=ci_data_fit.ci_pre_cti)

            fit = ci_fit.CIFit(masked_ci_data=ci_data_fit, cti_params=MockParams(),
                               cti_settings=MockSettings())

            assert (fit.residual_map == residual_map).all()

        def test__chi_squared_map_same_as_calculated_individually(self, ci_data_fit):
            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=ci_data_fit.image,
                                                                               mask=ci_data_fit.mask,
                                                                               model_data=ci_data_fit.ci_pre_cti)

            chi_sqaured_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map, noise_map=ci_data_fit.noise_map, mask=ci_data_fit.mask)

            fit = ci_fit.CIFit(masked_ci_data=ci_data_fit, cti_params=MockParams(),
                               cti_settings=MockSettings())

            assert (fit.chi_squared_map == chi_sqaured_map).all()

        def test__likelihood_same_as_calculated_individually(self, ci_data_fit):
            residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=ci_data_fit.image,
                                                                               mask=ci_data_fit.mask,
                                                                               model_data=ci_data_fit.ci_pre_cti)

            chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map, noise_map=ci_data_fit.noise_map, mask=ci_data_fit.mask)

            chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                             mask=ci_data_fit.mask)

            noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(noise_map=ci_data_fit.noise_map,
                                                                                       mask=ci_data_fit.mask)

            likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization)

            fit = ci_fit.CIFit(masked_ci_data=ci_data_fit, cti_params=MockParams(),
                               cti_settings=MockSettings())

            assert fit.likelihood == likelihood
            assert fit.figure_of_merit == fit.likelihood


class TestCIHyperFit:

    def test__image_and_ci_post_cti_the_same__noise_scaling_alls__likelihood_is_noise_normalization(self):
        ci_image = 10.0 * np.ones((2, 2))
        ci_noise_map = 2.0 * np.ones((2, 2))
        ci_mask = np.ma.zeros((2, 2))
        ci_pre_cti = 10.0 * np.ones((2, 2))
        ci_noise_scalings = [np.array([[0.0, 0.0], [0.0, 0.0]])]

        ci_data_hyper_fit = ci_data.MaskedCIHyperData(
            ci_frame=MockCIFrame(value=10.0), ci_pattern=MockPattern(), image=ci_image, noise_map=ci_noise_map,
            ci_pre_cti=ci_pre_cti, mask=ci_mask, noise_scaling_maps=ci_noise_scalings)

        hyper_noise_scalar = ci_hyper.CIHyperNoiseScalar(scale_factor=1.0)

        fit = ci_fit.CIHyperFit(
            masked_hyper_ci_data=ci_data_hyper_fit, cti_params=MockParams(), cti_settings=MockSettings(),
            hyper_noise_scalars=[hyper_noise_scalar])

        chi_squared = 0
        noise_normalization = 4.0 * np.log(2 * np.pi * 4.0)

        assert fit.likelihood == -0.5 * (chi_squared + noise_normalization)

    def test__image_and_post_cti_different__noise_scaling_alls__likelihood_chi_squared_and_noise_normalization(self):
        ci_image = 9.0 * np.ones((2, 2))
        ci_noise_map = 2.0 * np.ones((2, 2))
        ci_mask = np.ma.zeros((2, 2))
        ci_pre_cti = 10.0 * np.ones((2, 2))
        ci_noise_scalings = [np.array([[0.0, 0.0], [0.0, 0.0]])]

        ci_data_hyper_fit = ci_data.MaskedCIHyperData(
            ci_frame=MockCIFrame(value=10.0), ci_pattern=MockPattern(), image=ci_image, noise_map=ci_noise_map,
            ci_pre_cti=ci_pre_cti, mask=ci_mask, noise_scaling_maps=ci_noise_scalings)

        hyper_noise_scalar = ci_hyper.CIHyperNoiseScalar(scale_factor=1.0)

        fit = ci_fit.CIHyperFit(
            masked_hyper_ci_data=ci_data_hyper_fit, cti_params=MockParams(), cti_settings=MockSettings(),
            hyper_noise_scalars=[hyper_noise_scalar])

        chi_squared = 4.0 * ((1.0 / 2.0) ** 2.0)
        noise_normalization = 4.0 * np.log(2 * np.pi * 4.0)

        assert fit.likelihood == pytest.approx(-0.5 * (chi_squared + noise_normalization), 1e-4)

    def test__image_and_ci_post_cti_the_same__noise_scaling_nons__likelihood_is_noise_normalization(self):
        ci_image = 10.0 * np.ones((2, 2))
        ci_noise_map = 2.0 * np.ones((2, 2))
        ci_mask = np.ma.zeros((2, 2))
        ci_pre_cti = MockCIPreCTI(frame_geometry=MockGeometry(), ci_pattern=MockPattern(), array=10.0 * np.ones((2, 2)),
                                  value=10.0)

        ci_noise_scalings = [np.array([[1.0, 2.0], [3.0, 4.0]]),
                             np.array([[5.0, 6.0], [7.0, 8.0]])]

        ci_data_hyper_fit = ci_data.MaskedCIHyperData(
            ci_frame=MockCIFrame(value=10.0), ci_pattern=MockPattern(), image=ci_image, noise_map=ci_noise_map,
            ci_pre_cti=ci_pre_cti, mask=ci_mask, noise_scaling_maps=ci_noise_scalings)

        hyper_noise_scalar = ci_hyper.CIHyperNoiseScalar(scale_factor=1.0)

        fit = ci_fit.CIHyperFit(
            masked_hyper_ci_data=ci_data_hyper_fit, cti_params=MockParams(), cti_settings=MockSettings(),
            hyper_noise_scalars=[hyper_noise_scalar])

        chi_squared = 0
        noise_normalization = np.log(2 * np.pi * (2.0 + 1.0) ** 2.0) + np.log(2 * np.pi * (2.0 + 2.0) ** 2.0) + \
                              np.log(2 * np.pi * (2.0 + 3.0) ** 2.0) + np.log(2 * np.pi * (2.0 + 4.0) ** 2.0)

        assert fit.likelihood == -0.5 * (chi_squared + noise_normalization)

    def test__x2_noise_map_scaling_and_hyper_params__noise_map_term_comes_out_correct(self):
        ci_image = 10.0 * np.ones((2, 2))
        ci_noise_map = 3.0 * np.ones((2, 2))
        ci_mask = np.ma.zeros((2, 2))
        ci_pre_cti = MockCIPreCTI(frame_geometry=MockGeometry(), ci_pattern=MockPattern(), array=10.0 * np.ones((2, 2)),
                                  value=10.0)

        ci_noise_scalings = [np.array([[1.0, 2.0], [3.0, 4.0]]),
                             np.array([[5.0, 6.0], [7.0, 8.0]])]

        ci_data_hyper_fit = ci_data.MaskedCIHyperData(
            ci_frame=MockCIFrame(value=10.0), ci_pattern=MockPattern(), image=ci_image, noise_map=ci_noise_map,
            ci_pre_cti=ci_pre_cti, mask=ci_mask, noise_scaling_maps=ci_noise_scalings)

        hyper_noise_scalar_0 = ci_hyper.CIHyperNoiseScalar(scale_factor=1.0)
        hyper_noise_scalar_1 = ci_hyper.CIHyperNoiseScalar(scale_factor=2.0)

        fit = ci_fit.CIHyperFit(
            masked_hyper_ci_data=ci_data_hyper_fit, cti_params=MockParams(), cti_settings=MockSettings(),
            hyper_noise_scalars=[hyper_noise_scalar_0, hyper_noise_scalar_1])

        chi_squared = 0
        noise_normalization = np.log(2 * np.pi * (3.0 + 1.0 + 10.0) ** 2.0) + np.log(
            2 * np.pi * (3.0 + 2.0 + 12.0) ** 2.0) + \
                              np.log(2 * np.pi * (3.0 + 3.0 + 14.0) ** 2.0) + np.log(
            2 * np.pi * (3.0 + 4.0 + 16.0) ** 2.0)

        assert fit.likelihood == -0.5 * (chi_squared + noise_normalization)

    def test__all_quantities_are_same_as_calculated_individually(self, ci_data_hyper_fit):
        hyper_noise_scalar_0 = ci_hyper.CIHyperNoiseScalar(scale_factor=1.0)
        hyper_noise_scalar_1 = ci_hyper.CIHyperNoiseScalar(scale_factor=2.0)

        fit = ci_fit.CIHyperFit(
            masked_hyper_ci_data=ci_data_hyper_fit, cti_params=MockParams(), cti_settings=MockSettings(),
            hyper_noise_scalars=[hyper_noise_scalar_0, hyper_noise_scalar_1])

        hyper_noise_map = ci_fit.hyper_noise_map_from_noise_map_and_noise_scalings(
            noise_map=ci_data_hyper_fit.noise_map, noise_scaling_maps=ci_data_hyper_fit.noise_scaling_maps,
            hyper_noise_scalars=[hyper_noise_scalar_0, hyper_noise_scalar_1])

        assert (hyper_noise_map == fit.noise_map).all()

        residual_map = fit_util.residual_map_from_data_mask_and_model_data(data=ci_data_hyper_fit.image,
                                                                           mask=ci_data_hyper_fit.mask,
                                                                           model_data=ci_data_hyper_fit.ci_pre_cti)

        assert (residual_map == fit.residual_map).all()

        chi_squared_map = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map, noise_map=hyper_noise_map, mask=ci_data_hyper_fit.mask)

        assert (chi_squared_map == fit.chi_squared_map).all()

        chi_squared = fit_util.chi_squared_from_chi_squared_map_and_mask(chi_squared_map=chi_squared_map,
                                                                         mask=ci_data_hyper_fit.mask)

        noise_normalization = fit_util.noise_normalization_from_noise_map_and_mask(
            noise_map=hyper_noise_map, mask=ci_data_hyper_fit.mask)

        likelihood = fit_util.likelihood_from_chi_squared_and_noise_normalization(chi_squared=chi_squared,
                                                                                  noise_normalization=noise_normalization)

        assert (likelihood == fit.likelihood).all()


class TestScaledNoiseMap:

    def test__image_and_pre_cti_not_identical__noise_scalings_ares__no_noise_map_scaling(self):
        noise_map = 2.0 * np.ones((2, 2))
        noise_scaling_maps = [np.array([[0.0, 0.0], [0.0, 0.0]])]
        hyper_noises = [ci_hyper.CIHyperNoiseScalar(scale_factor=1.0)]

        noise_map = ci_fit.hyper_noise_map_from_noise_map_and_noise_scalings(noise_map=noise_map,
                                                                             noise_scaling_maps=noise_scaling_maps,
                                                                             hyper_noise_scalars=hyper_noises)

        assert (noise_map == (np.array([[2.0, 2.0],
                                        [2.0, 2.0]]))).all()

    def test__image_and_pre_cti_not_identical__factor_is__no_noise_map_scaling(self):
        noise_map = 2.0 * np.ones((2, 2))
        noise_scaling_maps = [np.array([[1.0, 2.0], [3.0, 4.0]])]
        hyper_noises = [ci_hyper.CIHyperNoiseScalar(scale_factor=0.0)]

        noise_map = ci_fit.hyper_noise_map_from_noise_map_and_noise_scalings(noise_map=noise_map,
                                                                             noise_scaling_maps=noise_scaling_maps,
                                                                             hyper_noise_scalars=hyper_noises)

        assert (noise_map == (np.array([[2.0, 2.0],
                                        [2.0, 2.0]]))).all()

    def test__image_and_pre_cti_not_identical__chi_sq_is_by_noise_map(self):
        noise_map = 2.0 * np.ones((2, 2))
        noise_scaling_maps = [np.array([[1.0, 2.0], [3.0, 4.0]])]
        hyper_noises = [ci_hyper.CIHyperNoiseScalar(scale_factor=1.0)]

        noise_map = ci_fit.hyper_noise_map_from_noise_map_and_noise_scalings(noise_map=noise_map,
                                                                             noise_scaling_maps=noise_scaling_maps,
                                                                             hyper_noise_scalars=hyper_noises)

        assert (noise_map == (np.array([[3.0, 4.0],
                                        [5.0, 6.0]]))).all()

    def test__x2_noise_map_scaling_and_hyper_params__noise_map_term_comes_out_correct(self):
        noise_map = 2.0 * np.ones((2, 2))
        noise_scaling_maps = [np.array([[1.0, 2.0], [3.0, 4.0]]), np.array([[1.0, 2.0], [3.0, 4.0]])]
        hyper_noises = [ci_hyper.CIHyperNoiseScalar(scale_factor=1.0), ci_hyper.CIHyperNoiseScalar(scale_factor=2.0)]

        noise_map = ci_fit.hyper_noise_map_from_noise_map_and_noise_scalings(noise_map=noise_map,
                                                                             noise_scaling_maps=noise_scaling_maps,
                                                                             hyper_noise_scalars=hyper_noises)

        assert (noise_map == (np.array([[5.0, 8.0],
                                        [11.0, 14.0]]))).all()


class MockCIDataFit(object):

    def __init__(self, image, noise_map, ci_pre_cti, mask, ci_pattern, ci_frame):
        self.image = image
        self.noise_map = noise_map
        self.ci_pre_cti = ci_pre_cti
        self.mask = mask
        self.ci_pattern = ci_pattern
        self.ci_frame = ci_frame
        self.is_hyper_data = False

    @property
    def chinj(self):
        return MockCIFrame(value=1.0)


@pytest.fixture(name='ci_data_fit_1')
def make_ci_data_fit_1():
    ci_image = np.array([[np.sqrt(1.0), np.sqrt(2.0)],
                         [np.sqrt(3.0), np.sqrt(4.0)]])
    ci_noise_map = 1.0 * np.ones((2, 2))
    ci_pattern = MockPattern(regions=[ci_frame.Region(region=[0, 1, 0, 1])])

    ci_pre_cti = MockCIPreCTI(frame_geometry=MockGeometry(),
                              ci_pattern=ci_pattern,
                              array=0.0 * np.ones((2, 2)), value=0.0)
    ci_mask = np.ma.zeros((2, 2))

    return MockCIDataFit(ci_frame=MockCIFrame(value=0.0), ci_pattern=ci_pattern, image=ci_image,
                         noise_map=ci_noise_map, ci_pre_cti=ci_pre_cti, mask=ci_mask)


class TestNoiseScalingMaps:

    def test__noise_scaling_map_of_ci_regions__extracts_correctly_from_chi_squard_map(self, ci_data_fit_1):
        # For the mock object MockCIFrame, the ci_regions_from_array function extracts the array entries
        # [0:2,0]

        fit = ci_fit.CIFit(masked_ci_data=ci_data_fit_1,
                           cti_params=MockParams(),
                           cti_settings=MockSettings())

        assert (fit.noise_scaling_map_of_ci_regions == fit.chi_squared_map[0:2, 0]).all()

    def test__noise_scaling_map_of_parallel_non_ci_regions__extracts_correctly_from_chi_squard_map(self,
                                                                                                   ci_data_fit_1):
        # For the mock object MockCIFrame, the parallel_non_ci_regions_frame_from_frame function extracts the array
        # entries [0:2,1]

        fit = ci_fit.CIFit(masked_ci_data=ci_data_fit_1,
                           cti_params=MockParams(),
                           cti_settings=MockSettings())

        assert (fit.noise_scaling_map_of_parallel_trails == fit.chi_squared_map[0:2, 1]).all()

    def test__noise_scaling_map_of_serial_trails__extracts_correctly_from_chi_squard_map(self,
                                                                                         ci_data_fit_1):
        # For the mock object MockCIFrame, the parallel_non_ci_regions_frame_from_frame function extracts the array
        # entries [0, 0:2]

        fit = ci_fit.CIFit(masked_ci_data=ci_data_fit_1,
                           cti_params=MockParams(),
                           cti_settings=MockSettings())

        assert (fit.noise_scaling_map_of_serial_trails == fit.chi_squared_map[0, 0:2]).all()

    def test__noise_scaling_map_of_overscan_above_serial_trails__extracts_correctly_from_chi_squard_map(self,
                                                                                                        ci_data_fit_1):
        # For the mock object MockCIFrame, the parallel_non_ci_regions_frame_from_frame function extracts the array
        # entries [0, 0:2]

        fit = ci_fit.CIFit(masked_ci_data=ci_data_fit_1,
                           cti_params=MockParams(),
                           cti_settings=MockSettings())

        assert (fit.noise_scaling_map_of_serial_overscan_above_trails == fit.chi_squared_map[1, 0:2]).all()



import os
import shutil

from autocti.charge_injection import ci_data, ci_frame, ci_pattern
from autocti.data import util

dirpath = os.path.dirname(os.path.realpath(__file__))


def reset_paths(test_name, output_path):
    try:
        shutil.rmtree(dirpath + '/data/' + test_name)
    except FileNotFoundError:
        pass

    try:
        shutil.rmtree(output_path + '/' + test_name)
    except FileNotFoundError:
        pass


class CIQuadGeometryIntegration(ci_frame.FrameGeometry):

    def __init__(self):
        """This class represents the quadrant geometry of an integration quadrant."""
        super(CIQuadGeometryIntegration, self).__init__(parallel_overscan=ci_frame.Region((33, 36, 1, 30)),
                                                        serial_overscan=ci_frame.Region((0, 33, 31, 36)),
                                                        serial_prescan=ci_frame.Region((0, 36, 0, 1)), corner=(0, 0))

    def rotate_for_parallel_cti(self, image):
        """ Rotate the quadrant image data before clocking via cti_settings in the parallel direction.

        For the integration quadrant, no rotation is required for parallel clocking

        Params
        ----------
        image_pre_clocking : ndarray
            The ci_pre_ctis before parallel clocking, therefore before it has been reoriented for clocking.
        """
        return image

    def rotate_before_serial_cti(self, image_pre_clocking):
        """ Rotate the quadrant image data before clocking via cti_settings in the serial direction.

        For the integration quadrant, the image is rotated 90 degrees anti-clockwise for serial clocking.

        NOTE : The NumPy transpose routine does not reorder the array's memory, making it non-contiguous. This is not \
        a useable data-type for C++ (and therefore cti_settings), so we use .copy() to force a memory re-ordering.

        Params
        ----------
        image_pre_clocking : ndarray
            The ci_pre_ctis before serial clocking, therefore before it has been reoriented for clocking.
        """

        return image_pre_clocking.T.copy()

    def rotate_after_serial_cti(self, image_post_clocking):
        """ Re-rotate the quadrant image data after clocking via cti_settings in the serial direction.

        For the integration quadrant, the ci_pre_ctis is re-rotated 90 degrees clockwise after serial clocking.


        NOTE : The NumPy transpose routine does not reorder the array's memory, making it non-contiguous. This is not \
        a useable data-type for C++ (and therefore cti_settings), so we use .copy() to force a memory re-ordering.

        Params
        ----------
        image_post_clocking : ndarray
            The ci_pre_ctis after clocking, therefore with serial cti added or corrected.

        """
        return image_post_clocking.T.copy()

    def parallel_trail_from_y(self, y, dy):
        """Coordinates of a parallel trail of size dy from coordinate y"""
        return y, y + dy + 1

    def serial_trail_from_x(self, x, dx):
        """Coordinates of a serial trail of size dx from coordinate x"""
        return x, x + dx + 1


shape = (36, 36)
ci_regions = [(1, 7, 3, 30), (17, 23, 3, 30)]
frame_geometry = CIQuadGeometryIntegration()


def simulate_integration_quadrant(test_name, normalizations, cti_params, cti_settings):
    output_path = "{}/data/".format(os.path.dirname(os.path.realpath(__file__))) + test_name + '/'

    if not os.path.exists(output_path):
        os.makedirs(output_path)

    sim_ci_patterns = ci_pattern.uniform_from_lists(normalizations=normalizations, regions=ci_regions)

    sim_ci_pre_ctis = list(map(lambda ci_pattern : ci_pattern.simulate_ci_pre_cti(shape=shape), sim_ci_patterns))

    sim_ci_datas = list(map(lambda ci_pre_cti, pattern:
                            ci_data.simulate(ci_pre_cti=ci_pre_cti, frame_geometry=frame_geometry,
                                             ci_pattern=pattern, cti_settings=cti_settings,
                                             cti_params=cti_params,
                                             read_noise=None),
                            sim_ci_pre_ctis, sim_ci_patterns))

    list(map(lambda sim_ci_data, index:
             util.numpy_array_to_fits(array=sim_ci_data.image, file_path=output_path + '/ci_image_' + str(index) + '.fits'),
             sim_ci_datas, range(len(sim_ci_datas))))



import numpy as np

from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_fit
from autocti.charge_injection.plotters import fit_plotters
from autocti.data import mask as msk
from autocti.model import arctic_params
from autocti.model import arctic_settings
from test.charge_injection.plotters.fixtures import *
from test.mock.mock import MockPattern, MockCIFrame


@pytest.fixture(name='fit_plotter_path')
def make_ci_data_plotter_setup():
    return "{}/../../test_files/plotting/fit/".format(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture(name='mask')
def make_mask():
    return msk.Mask.empty_for_shape(shape=(2, 2))


@pytest.fixture(name='image')
def make_image():
    return np.ones((2, 2))


@pytest.fixture(name='noise_map')
def make_noise_map():
    return 2.0 * np.ones((2, 2))


@pytest.fixture(name='ci_pre_cti')
def make_ci_pre_cti():
    return 3.0 * np.ones((2, 2))


@pytest.fixture(name='ci_data_fit')
def make_ci_data_fit(image, noise_map, mask, ci_pre_cti):
    return ci_data.MaskedCIData(image=image, noise_map=noise_map, ci_pre_cti=ci_pre_cti, mask=mask,
                                ci_pattern=MockPattern(), ci_frame=MockCIFrame(value=3.0))


@pytest.fixture(name="cti_settings")
def make_cti_settings():
    parallel_settings = arctic_settings.Settings(well_depth=0, niter=1, express=1, n_levels=2000)
    return arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings)


@pytest.fixture(name="cti_params")
def make_cti_params():
    parallel_1_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
    return arctic_params.ArcticParams(parallel_species=parallel_1_species)


@pytest.fixture(name="fit")
def make_fit(ci_data_fit, cti_params, cti_settings):
    return ci_fit.CIFit(masked_ci_data=ci_data_fit, cti_params=cti_params,
                                                               cti_settings=cti_settings)


def test__image_is_output(fit, mask, fit_plotter_path, plot_patch):
    fit_plotters.plot_image(fit=fit, mask=mask, extract_array_from_mask=True,
                            cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                            output_path=fit_plotter_path, output_format='png')
    assert fit_plotter_path + 'fit_image.png' in plot_patch.paths


def test__noise_map_is_output(fit, mask, fit_plotter_path, plot_patch):
    fit_plotters.plot_noise_map(fit=fit, mask=mask, extract_array_from_mask=True,
                                cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                output_path=fit_plotter_path, output_format='png')
    assert fit_plotter_path + 'fit_noise_map.png' in plot_patch.paths


def test__signal_to_noise_map_is_output(fit, mask, fit_plotter_path, plot_patch):
    fit_plotters.plot_signal_to_noise_map(fit=fit, mask=mask, extract_array_from_mask=True,
                                          cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                          output_path=fit_plotter_path, output_format='png')
    assert fit_plotter_path + 'fit_signal_to_noise_map.png' in plot_patch.paths


def test__ci_pre_cti_is_output(fit, mask, fit_plotter_path, plot_patch):
    fit_plotters.plot_ci_pre_cti(fit=fit, mask=mask, extract_array_from_mask=True,
                                 cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                 output_path=fit_plotter_path, output_format='png')
    assert fit_plotter_path + 'fit_ci_pre_cti.png' in plot_patch.paths


def test__ci_post_cti_is_output(fit, mask, fit_plotter_path, plot_patch):
    fit_plotters.plot_ci_post_cti(fit=fit, mask=mask, extract_array_from_mask=True,
                                  cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                  output_path=fit_plotter_path, output_format='png')
    assert fit_plotter_path + 'fit_ci_post_cti.png' in plot_patch.paths


def test__residual_map_is_output(fit, mask, fit_plotter_path, plot_patch):
    fit_plotters.plot_residual_map(fit=fit, mask=mask, extract_array_from_mask=True,
                                   cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                   output_path=fit_plotter_path, output_format='png')
    assert fit_plotter_path + 'fit_residual_map.png' in plot_patch.paths


def test__chi_squared_map_is_output(fit, mask, fit_plotter_path, plot_patch):
    fit_plotters.plot_chi_squared_map(fit=fit, mask=mask, extract_array_from_mask=True,
                                      cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                      output_path=fit_plotter_path, output_format='png')
    assert fit_plotter_path + 'fit_chi_squared_map.png' in plot_patch.paths



import matplotlib.pyplot as plt
from astropy.io import fits

from autocti import exc
from autocti.data import util

def get_subplot_rows_columns_figsize(number_subplots):
    """Get the size of a sub plot in (rows, columns), based on the number of subplots that are going to be plotted.

    Parameters
    -----------
    number_subplots : int
        The number of subplots that are to be plotted in the figure.
    """
    if number_subplots <= 2:
        return 1, 2, (18, 8)
    elif number_subplots <= 4:
        return 2, 2, (13, 10)
    elif number_subplots <= 6:
        return 2, 3, (18, 12)
    elif number_subplots <= 9:
        return 3, 3, (25, 20)
    elif number_subplots <= 12:
        return 3, 4, (25, 20)
    elif number_subplots <= 16:
        return 4, 4, (25, 20)
    elif number_subplots <= 20:
        return 4, 5, (25, 20)
    else:
        return 6, 6, (25, 20)


def setup_figure(figsize, as_subplot):
    """Setup a figure for plotting an hyper.

    Parameters
    -----------
    figsize : (int, int)
        The size of the figure in (rows, columns).
    as_subplot : bool
        If the figure is a subplot, the setup_figure function is omitted to ensure that each subplot does not create a \
        new figure and so that it can be output using the *output_subplot_array* function.
    """
    if not as_subplot:
        plt.figure(figsize=figsize)


def set_title(title, titlesize):
    """Set the title and title size of the figure.

    Parameters
    -----------
    title : str
        The text of the title.
    titlesize : int
        The size of of the title of the figure.
    """
    plt.title(title, fontsize=titlesize)


def output_figure(array, as_subplot, output_path, output_filename, output_format):
    """Output the figure, either as an hyper on the screen or to the hard-disk as a .png or .fits file.

    Parameters
    -----------
    array : ndarray
        The 2D array of hyper to be output, required for outputting the hyper as a fits file.
    as_subplot : bool
        Whether the figure is part of subplot, in which case the figure is not output so that the entire subplot can \
        be output instead using the *output_subplot_array* function.
    output_path : str
        The path on the hard-disk where the figure is output.
    output_filename : str
        The filename of the figure that is output.
    output_format : str
        The format the figue is output:
        'show' - display on computer screen.
        'png' - output to hard-disk as a png.
        'fits' - output to hard-disk as a fits file.'
    """
    if not as_subplot:

        if output_format is 'show':
            plt.show()
        elif output_format is 'png':
            plt.savefig(output_path + output_filename + '.png', bbox_inches='tight')
        elif output_format is 'fits':
            util.numpy_array_to_fits(array=array, file_path=output_path + output_filename + '.fits', overwrite=True)


def output_subplot_array(output_path, output_filename, output_format):
    """Output a figure which consists of a set of subplot,, either as an hyper on the screen or to the hard-disk as a \
    .png file.

    Parameters
    -----------
    output_path : str
        The path on the hard-disk where the figure is output.
    output_filename : str
        The filename of the figure that is output.
    output_format : str
        The format the figue is output:
        'show' - display on computer screen.
        'png' - output to hard-disk as a png.
    """
    if output_format is 'show':
        plt.show()
    elif output_format is 'png':
        plt.savefig(output_path + output_filename + '.png', bbox_inches='tight')
    elif output_format is 'fits':
        raise exc.PlottingException('You cannot output a subplots with format .fits')


def close_figure(as_subplot):
    """After plotting and outputting a figure, close the matplotlib figure instance (omit if a subplot).

    Parameters
    -----------
    as_subplot : bool
        Whether the figure is part of subplot, in which case the figure is not closed so that the entire figure can \
        be closed later after output.
    """
    if not as_subplot:
        plt.close()



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: python/VIS_CTI_ChargeInjection/CIImage.pyy

Created on: 02/14/18
Author: James Nightingale
"""

import numpy as np

from autocti.charge_injection import ci_frame


class Mask(np.ndarray):

    def __new__(cls, array, *args, **kwargs):
        mask = np.array(array, dtype='float64').view(cls)
        return mask

    @classmethod
    def empty_for_shape(cls, shape):
        """
        Create the mask used for CTI Calibration as all False's (e.g. no masking).

        Parameters
        ----------
        shape : (int, int)
            The dimensions of the 2D mask.
        """
        # noinspection PyArgumentList
        return cls(array=np.full(shape, False))

    @classmethod
    def create(cls, shape, frame_geometry, regions=None, cosmic_rays=None, cr_parallel=0, cr_serial=0,
               cr_diagonal=0):
        """
        Create the mask used for CTI Calibration, which is all False unless specific regions are input for masking.

        Parameters
        ----------
        cr_diagonal
        shape : (int, int)
            The dimensions of the 2D mask.
        frame_geometry : ci_frame.CIQuadGeometry
            The quadrant geometry of the simulated image, defining where the parallel / serial overscans are and \
            therefore the direction of clocking and rotations before input into the cti algorithm.
        regions : [(int,)]
            A list of the regions on the image to mask.
        cosmic_rays : ndarray.ma
            2D array flagging where cosmic rays on the image.
        cr_parallel : int
            If a cosmic-ray mask is supplied, the number of pixels from each ray pixels are masked in the parallel \
            direction.
        cr_serial : int
            If a cosmic-ray mask is supplied, the number of pixels from each ray pixels are masked in the serial \
            direction.
        """
        mask = cls.empty_for_shape(shape)

        if regions is not None:
            mask.regions = list(map(lambda r: ci_frame.Region(r), regions))
            for region in mask.regions:
                mask[region.y0:region.y1, region.x0:region.x1] = True
        elif regions is None:
            mask.regions = None

        if cosmic_rays is not None:
            for y in range(mask.shape[0]):
                for x in range(mask.shape[1]):
                    if cosmic_rays[y, x]:
                        y0, y1 = frame_geometry.parallel_trail_from_y(y, cr_parallel)
                        mask[y0:y1, x] = True
                        x0, x1 = frame_geometry.serial_trail_from_x(x, cr_serial)
                        mask[y, x0:x1] = True
                        y0, y1 = frame_geometry.parallel_trail_from_y(y, cr_diagonal)
                        x0, x1 = frame_geometry.serial_trail_from_x(x, cr_diagonal)
                        mask[y0:y1, x0:x1] = True

        elif cosmic_rays is None:
            mask.cosmic_rays = None

        return mask



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: tests/python/PyArCTIC_test.py

Created on: 02/13/18
Author: James Nightingale
"""

import numpy as np
import pytest

from autocti.model import arctic_params
from autocti.model import arctic_settings
from autocti.model import pyarctic


def setup(include_parallel=False, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
          p_well_notch_depth=0.01, p_well_fill_alpha=1.0, p_well_fill_beta=0.8,
          p_well_fill_gamma=0.0, include_serial=False, s_trap_densities=(0.05, 0.05,),
          s_trap_lifetimes=(1.0, 1.0),
          s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
          s_well_fill_gamma=0.0):
    serial_species = []
    parallel_species = []
    serial_ccd = None
    parallel_ccd = None

    if include_serial:
        serial_ccd = arctic_params.CCD(well_notch_depth=s_well_notch_depth, well_fill_alpha=s_well_fill_alpha,
                                       well_fill_beta=s_well_fill_beta, well_fill_gamma=s_well_fill_gamma)
        serial_species = [arctic_params.Species(trap_density=s_trap_densities[i], trap_lifetime=s_trap_lifetimes[i]) for
                          i in range(len(s_trap_densities))]
    if include_parallel:
        parallel_ccd = arctic_params.CCD(well_notch_depth=p_well_notch_depth, well_fill_alpha=p_well_fill_alpha,
                                         well_fill_beta=p_well_fill_beta, well_fill_gamma=p_well_fill_gamma)
        parallel_species = [arctic_params.Species(trap_density=p_trap_densities[i], trap_lifetime=p_trap_lifetimes[i])
                            for
                            i in range(len(p_trap_densities))]

    return arctic_params.ArcticParams(parallel_ccd=parallel_ccd, serial_ccd=serial_ccd, serial_species=serial_species,
                                      parallel_species=parallel_species)


def setup_settings(include_parallel=False, p_well_depth=84700, p_niter=1, p_express=5, p_n_levels=2000,
                   p_charge_injection_mode=False,
                   p_readout_offset=0,
                   include_serial=False, s_well_depth=84700, s_niter=1, s_express=5, s_n_levels=2000,
                   s_charge_injection_mode=False,
                   s_readout_offset=0):
    """Factory to set up a *ParallelParams* and / or *SerialParams* sub-class as an *ArcticParams*
    instance using any number of trap species in both directions.

    Parameters
    ----------
    s_n_levels
    s_express
    s_niter
    s_well_depth
    s_readout_offset
    s_charge_injection_mode
    include_parallel: Bool
        If True parallel parameters will be included in the ArcticParams object
    include_serial: Bool
        If True serial parameters will be included in the ArcticParams object
    p_well_depth : int
        The full well depth of the CCD.
    p_niter : int
        If CTI is being corrected, niter determines the number of times clocking is run to erform the \
        correction via forward modeling. For adding CTI only one run is required and niter is ignored.
    p_express : int
        The factor by which pixel-to-pixel transfers are combined for efficiency.
    p_n_levels : int
        Relic of old arctic code, not used anymore and will be removed in future.
    p_charge_injection_mode : bool
        If True, clocking is performed in charge injection line mode, whereby each pixel is clocked and therefore \
         trailed by traps over the entire CCD (as opposed to its distance from the CCD register).
    p_readout_offset : int
        Introduces an offset which increases the number of transfers each pixel takes in the parallel direction.
    """

    parallel_settings = arctic_settings.Settings(p_well_depth, p_niter, p_express, p_n_levels,
                                                 p_charge_injection_mode,
                                                 p_readout_offset) if include_parallel else None

    serial_settings = arctic_settings.Settings(s_well_depth, s_niter, s_express, s_n_levels,
                                               s_charge_injection_mode, s_readout_offset) if include_serial else None

    return arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings, serial=serial_settings)


def add_parallel_cti_to_image(image, params, settings):
    return pyarctic.call_arctic(image, params.parallel_species, params.parallel_ccd, settings.parallel)


def add_serial_cti_to_image(image, params, settings):
    return pyarctic.call_arctic(image, params.serial_species, params.serial_ccd, settings.serial)


def correct_parallel_cti_from_image(image, params, settings):
    return pyarctic.call_arctic(image, params.parallel_species, params.parallel_ccd, settings.parallel,
                                correct_cti=True)


def correct_serial_cti_from_image(image, params, settings):
    return pyarctic.call_arctic(image, params.serial_species, params.serial_ccd, settings.serial, correct_cti=True)


@pytest.fixture(scope='class', name='arctic_parallel')
def make_arctic_parallel():
    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                 charge_injection_mode=True, readout_offset=0)

    return arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings)


@pytest.fixture(scope='class', name='arctic_serial')
def make_arctic_serial():
    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                               readout_offset=0)
    return arctic_settings.ArcticSettings(neomode='NEO', serial=serial_settings)


@pytest.fixture(scope='class', name='arctic_both')
def make_arctic_both():
    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                 readout_offset=0)

    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                               readout_offset=0)
    return arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings,
                                          serial=serial_settings)


class TestArcticAddCTI:
    class TestParallelCTI:

        def test__horizontal_line__line_loses_charge_trails_appear(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_alpha=1.0, p_well_fill_beta=0.8,
                               p_well_fill_gamma=0.0)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_difference[2,
                    :] < 0.0).all()  # All pixels in the charge line should lose charge due to capture
            assert (image_difference[3:-1, :] > 0.0).all()  # All other pixels should have charge trailed into them

        def test__vertical_line__no_trails(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_alpha=1.0, p_well_fill_beta=0.8,
                               p_well_fill_gamma=0.0)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()  # Most pixels unchanged
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()  # charge line still loses charge

        def test__double_trap_density__more_captures_so_brighter_trails(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.01, p_well_fill_alpha=1.0, p_well_fill_beta=0.8,
                                 p_well_fill_gamma=0.0)

            parameters_1 = setup(include_parallel=True, p_trap_densities=(0.2,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.01, p_well_fill_alpha=1.0, p_well_fill_beta=0.8,
                                 p_well_fill_gamma=0.0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                         settings=arctic_parallel)
            image_post_cti_1 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                         settings=arctic_parallel)

            assert (image_post_cti_0[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_post_cti_1[0:2, :] == 0.0).all()  # First four rows should all remain zero
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[2, :] > image_post_cti_1[4, :]).all()  # charge line loses less charge in image 1
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[3:-1, :] < image_post_cti_1[5:-1, :]).all()
            # fewer pixels trailed behind in image 2

        def test__double_trap_lifetime__longer_release_so_fainter_trails(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.01, p_well_fill_alpha=1.0, p_well_fill_beta=0.8,
                                 p_well_fill_gamma=0.0)

            parameters_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(2.0,),
                                 p_well_notch_depth=0.01, p_well_fill_alpha=1.0, p_well_fill_beta=0.8,
                                 p_well_fill_gamma=0.0)
            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                         settings=arctic_parallel)
            image_post_cti_1 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                         settings=arctic_parallel)

            assert (image_post_cti_0[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_post_cti_1[0:2, :] == 0.0).all()  # First four rows should all remain zero
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[2, :] == image_post_cti_1[2, :]).all()  # charge line loses equal amount of charge
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[3:-1, :] > image_post_cti_1[3:-1, :]).all()
            # each trail in pixel 2 is 'longer' so fainter

        def test__increase_beta__fewer_captures_fainter_trail(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            parameters_1 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.01, p_well_fill_beta=0.9)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                         settings=arctic_parallel)
            image_post_cti_1 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                         settings=arctic_parallel)

            assert (image_post_cti_0[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_post_cti_1[0:2, :] == 0.0).all()  # First four rows should all remain zero
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[2, :] < image_post_cti_1[2, :]).all()
            # charge line loses less charge with higher beta
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[3:-1, :] > image_post_cti_1[3:-1, :]).all()  # so less electrons trailed into image

        def test__two_traps_half_density_of_one__same_trails(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            parameters_1 = setup(include_parallel=True, p_trap_densities=(0.05, 0.05),
                                 p_trap_lifetimes=(1.0, 1.0),
                                 p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                         settings=arctic_parallel)
            image_post_cti_1 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                         settings=arctic_parallel)
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0 == image_post_cti_1).all()

        def test__three_traps_third_density_of_one__same_trails(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_parallel=True, p_trap_densities=(0.15,), p_trap_lifetimes=(1.0,),
                                 p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            parameters_1 = setup(include_parallel=True, p_trap_densities=(0.05, 0.05, 0.05,),
                                 p_trap_lifetimes=(1.0, 1.0, 1.0),
                                 p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                         settings=arctic_parallel)
            image_post_cti_1 = add_parallel_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                         settings=arctic_parallel)
            image_difference = abs(image_post_cti_0 - image_post_cti_1)

            assert (image_difference < 1e-9).all()

        def test__delta_functions__add_cti_only_behind_them(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

    class TestSerialCTI:

        def test__horizontal_line__line_loses_charge_trails_appear(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            parameters = setup(include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                               s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                               s_well_fill_gamma=0.0)

            image_post_cti = add_serial_cti_to_image(image=image_pre_cti, params=parameters,
                                                     settings=arctic_serial)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_difference[2,
                    :] < 0.0).all()  # All pixels in the charge line should lose charge due to capture
            assert (image_difference[3:-1, :] > 0.0).all()  # All other pixels should have charge trailed into them

        def test__vertical_line__no_trails(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                               s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                               s_well_fill_gamma=0.0)

            image_post_cti = add_serial_cti_to_image(image=image_pre_cti, params=parameters,
                                                     settings=arctic_serial)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()  # Most pixels unchaged
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()  # charge line still loses charge

        def test__double_trap_density__more_captures_so_brighter_trails(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                                 s_well_fill_gamma=0.0)

            parameters_1 = setup(include_serial=True, s_trap_densities=(0.2,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                                 s_well_fill_gamma=0.0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                       settings=arctic_serial)
            image_post_cti_1 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                       settings=arctic_serial)

            assert (image_post_cti_0[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_post_cti_1[0:2, :] == 0.0).all()  # First four rows should all remain zero
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[2, :] > image_post_cti_1[4, :]).all()  # charge line loses less charge in image 1
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[3:-1, :] < image_post_cti_1[5:-1, :]).all()
            # fewer pixels trailed behind in image 2

        def test__double_trap_lifetime__longer_release_so_fainter_trails(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                                 s_well_fill_gamma=0.0)

            parameters_1 = setup(include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(2.0,),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                                 s_well_fill_gamma=0.0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                       settings=arctic_serial)
            image_post_cti_1 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                       settings=arctic_serial)

            assert (image_post_cti_0[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_post_cti_1[0:2, :] == 0.0).all()  # First four rows should all remain zero
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[2, :] == image_post_cti_1[2, :]).all()  # charge line loses equal amount of charge
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[3:-1, :] > image_post_cti_1[3:-1, :]).all()
            # each trail in pixel 2 is 'longer' so fainter

        def test__increase_beta__fewer_captures_fainter_trail(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                                 s_well_fill_gamma=0.0)

            parameters_1 = setup(include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.9,
                                 s_well_fill_gamma=0.0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                       settings=arctic_serial)
            image_post_cti_1 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                       settings=arctic_serial)

            assert (image_post_cti_0[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_post_cti_1[0:2, :] == 0.0).all()  # First four rows should all remain zero
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[2, :] < image_post_cti_1[2, :]).all()
            # charge line loses less charge with higher beta
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[3:-1, :] > image_post_cti_1[3:-1, :]).all()  # so less electrons trailed into image

        def test__two_traps_half_density_of_one__same_trails(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                                 s_well_fill_gamma=0.0)

            parameters_1 = setup(include_serial=True, s_trap_densities=(0.05, 0.05,),
                                 s_trap_lifetimes=(1.0, 1.0),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                                 s_well_fill_gamma=0.0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                       settings=arctic_serial)
            image_post_cti_1 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                       settings=arctic_serial)
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0 == image_post_cti_1).all()

        def test__three_traps_third_density_of_one__same_trails(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            parameters_0 = setup(include_serial=True, s_trap_densities=(0.15,), s_trap_lifetimes=(1.0,),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                                 s_well_fill_gamma=0.0)

            parameters_1 = setup(include_serial=True, s_trap_densities=(0.05, 0.05, 0.05,),
                                 s_trap_lifetimes=(1.0, 1.0, 1.0),
                                 s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                                 s_well_fill_gamma=0.0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_0,
                                                       settings=arctic_serial)
            image_post_cti_1 = add_serial_cti_to_image(image=image_pre_cti, params=parameters_1,
                                                       settings=arctic_serial)

            image_difference = abs(image_post_cti_0 - image_post_cti_1)

            assert (image_difference < 1e-9).all()

        def test__delta_functions__add_cti_only_behind_them(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_serial=True, s_trap_densities=(0.1,), s_trap_lifetimes=(1.0,),
                               s_well_notch_depth=0.01, s_well_fill_alpha=1.0, s_well_fill_beta=0.8,
                               s_well_fill_gamma=0.0)

            image_post_cti = add_serial_cti_to_image(image=image_pre_cti, params=parameters,
                                                     settings=arctic_serial)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

    class TestRectangularImageDimensions:

        def test__horizontal_line__rectangular_image_odd_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 7))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()
            assert (image_difference[2, :] < 0.0).all()
            assert (image_difference[3:-1, :] > 0.0).all()

        def test__horizontal_line__rectangular_image_even_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((4, 6))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()
            assert (image_difference[2, :] < 0.0).all()
            assert (image_difference[3:-1, :] > 0.0).all()

        def test__horizontal_line__rectangular_image_even_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((4, 7))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()
            assert (image_difference[2, :] < 0.0).all()
            assert (image_difference[3:-1, :] > 0.0).all()

        def test__horizontal_line__rectangular_image_odd_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 6))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()
            assert (image_difference[2, :] < 0.0).all()
            assert (image_difference[3:-1, :] > 0.0).all()

        def test__vertical_line__rectangular_image_odd_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((3, 5))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()

        def test__vertical_line__rectangular_image_even_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((4, 6))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()

        def test__vertical_line__rectangular_image_even_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((4, 7))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()

        def test__vertical_line__rectangular_image_odd_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 6))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()

        def test__delta_functions__add_cti_only_behind_them__odd_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 7))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

            assert (image_difference[:, 5] == 0.0).all()  # No Delta, no charge
            assert (image_difference[:, 6] == 0.0).all()  # No Delta, no charge

        def test__delta_functions__add_cti_only_behind_them__even_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((6, 8))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

            assert (image_difference[:, 5] == 0.0).all()  # No Delta, no charge
            assert (image_difference[:, 6] == 0.0).all()  # No Delta, no charge
            assert (image_difference[:, 7] == 0.0).all()  # No Delta, no charge

        def test__delta_functions__add_cti_only_behind_them__even_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((6, 7))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

            assert (image_difference[:, 5] == 0.0).all()  # No Delta, no charge
            assert (image_difference[:, 6] == 0.0).all()  # No Delta, no charge

        def test__delta_functions__add_cti_only_behind_them__odd_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 6))

            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

            assert (image_difference[:, 5] == 0.0).all()  # No Delta, no charge

    class TestChargeInjectionMode:

        def test__charge_injection_parameters_on__makes_trails_longer(self):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            settings = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=1, p_express=5,
                                      p_n_levels=2000,
                                      p_charge_injection_mode=False, p_readout_offset=0)

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=settings)

            settings_charge_inj = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=1,
                                                 p_express=5, p_n_levels=2000,
                                                 p_charge_injection_mode=True, p_readout_offset=0)

            image_post_cti_charge_inj = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                                  settings=settings_charge_inj)

            image_difference = image_post_cti_charge_inj - image_post_cti

            assert (image_difference[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_difference[2,
                    :] < 0.0).all()  # With charge injection mode on, more transfers, so more charge lost
            assert (image_difference[3:-1, :] > 0.0).all()  # And therefore, trails are brighter

    class TestDensityVary:

        def test__horizontal_line__different_density_in_each_column(self):
            settings = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=1, p_express=5,
                                      p_n_levels=2000,
                                      p_charge_injection_mode=False, p_readout_offset=0)

            ccd = arctic_params.CCD(well_fill_beta=0.8, well_notch_depth=0.01)
            species = [arctic_params.Species(trap_density=10., trap_lifetime=1.),
                       arctic_params.Species(trap_density=20., trap_lifetime=1.),
                       arctic_params.Species(trap_density=30., trap_lifetime=1.),
                       arctic_params.Species(trap_density=40., trap_lifetime=1.),
                       arctic_params.Species(trap_density=50., trap_lifetime=1.), ]

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            image_post_cti = pyarctic.call_arctic_parallel_density_vary(image=image_pre_cti, species=species, ccd=ccd,
                                                                        settings=settings.parallel)

            # noinspection PyUnresolvedReferences
            assert (image_post_cti[2:5, 0] != image_post_cti[2:5, 1]).all()
            # noinspection PyUnresolvedReferences
            assert (image_post_cti[2:5, 1] != image_post_cti[2:5, 2]).all()
            # noinspection PyUnresolvedReferences
            assert (image_post_cti[2:5, 2] != image_post_cti[2:5, 3]).all()
            # noinspection PyUnresolvedReferences
            assert (image_post_cti[2:5, 3] != image_post_cti[2:5, 4]).all()


class TestArcticCorrectCTI:
    class TestParallel:

        def test__horizontal_line__corrected_image_more_like_original(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__vertical_line__corrected_image_more_like_original(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__corrected_image_more_like_original(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__decrease_niter__worse_correction(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            arctic_niter_5 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=5, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_5)

            image_difference_niter_5 = image_correct_cti - image_pre_cti

            arctic_niter_3 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=3, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_3)

            image_difference_niter_3 = image_correct_cti - image_pre_cti

            # noinspection PyUnresolvedReferences
            assert (abs(image_difference_niter_5) <= abs(
                image_difference_niter_3)).all()  # First four rows should all remain zero

    class TestSerial:

        def test__horizontal_line__corrected_image_more_like_original(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] = 100

            parameters = setup(include_serial=True, s_trap_densities=(0.2,), s_trap_lifetimes=(2.0,),
                               s_well_notch_depth=0.02, s_well_fill_beta=0.4)

            image_post_cti = add_serial_cti_to_image(image=image_pre_cti, params=parameters,
                                                     settings=arctic_serial)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_serial_cti_from_image(image=image_post_cti, params=parameters,
                                                              settings=arctic_serial)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()

        def test__vertical_line__corrected_image_more_like_original(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_serial=True, s_trap_densities=(0.2,), s_trap_lifetimes=(2.0,),
                               s_well_notch_depth=0.02, s_well_fill_beta=0.4)

            image_post_cti = add_serial_cti_to_image(image=image_pre_cti, params=parameters,
                                                     settings=arctic_serial)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_serial_cti_from_image(image=image_post_cti, params=parameters,
                                                              settings=arctic_serial)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()

        def test__delta_functions__corrected_image_more_like_original(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_serial=True, s_trap_densities=(0.2,), s_trap_lifetimes=(2.0,),
                               s_well_notch_depth=0.02, s_well_fill_beta=0.4)

            image_post_cti = add_serial_cti_to_image(image=image_pre_cti, params=parameters,
                                                     settings=arctic_serial)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_serial_cti_from_image(image=image_post_cti, params=parameters,
                                                              settings=arctic_serial)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__decrease_niter__worse_correction(self, arctic_serial):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_serial=True, s_trap_densities=(0.2,), s_trap_lifetimes=(2.0,),
                               s_well_notch_depth=0.02, s_well_fill_beta=0.4)

            image_post_cti = add_serial_cti_to_image(image=image_pre_cti, params=parameters,
                                                     settings=arctic_serial)

            arctic_niter_5 = setup_settings(include_serial=True, s_well_depth=84700, s_niter=5, s_express=5,
                                            s_n_levels=2000,
                                            s_charge_injection_mode=False, s_readout_offset=0)

            image_correct_cti = correct_serial_cti_from_image(image=image_post_cti, params=parameters,
                                                              settings=arctic_niter_5)

            image_difference_niter_5 = image_correct_cti - image_pre_cti

            arctic_niter_3 = setup_settings(include_serial=True, s_well_depth=84700, s_niter=3, s_express=5,
                                            s_n_levels=2000,
                                            s_charge_injection_mode=False, s_readout_offset=0)

            image_correct_cti = correct_serial_cti_from_image(image=image_post_cti, params=parameters,
                                                              settings=arctic_niter_3)

            image_difference_niter_3 = image_correct_cti - image_pre_cti

            # noinspection PyUnresolvedReferences
            assert (abs(image_difference_niter_5) <= abs(
                image_difference_niter_3)).all()  # First four rows should all remain zero

    class TestRectangularImageDimensions:

        def test__horizontal_line__odd_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 3))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__horizontal_line__even_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((6, 4))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__horizontal_line__even_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((6, 3))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__horizontal_line__odd_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 4))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__veritcal_line__odd_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 3))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__veritcal_line__even_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((6, 4))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__veritcal_line__even_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((6, 3))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__veritcal_line__odd_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 4))
            image_pre_cti[:, 2] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__odd_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 7))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__even_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((6, 8))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__even_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((6, 7))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__odd_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 8))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_parallel)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__decrease_niter__worse_correction__odd_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((3, 5))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            arctic_niter_5 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=5, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_5)

            image_difference_niter_5 = image_correct_cti - image_pre_cti

            arctic_niter_3 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=3, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_3)

            image_difference_niter_3 = image_correct_cti - image_pre_cti

            # noinspection PyUnresolvedReferences
            assert (abs(image_difference_niter_5) <= abs(
                image_difference_niter_3)).all()  # First four rows should all remain zero

        def test__dencrease_niter__worse_correction__even_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((4, 4))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            arctic_niter_5 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=5, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_5)

            image_difference_niter_5 = image_correct_cti - image_pre_cti

            arctic_niter_3 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=3, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_3)

            image_difference_niter_3 = image_correct_cti - image_pre_cti

            # noinspection PyUnresolvedReferences
            assert (abs(image_difference_niter_5) <= abs(
                image_difference_niter_3)).all()  # First four rows should all remain zero

        def test__dencrease_niter__worse_correction__even_x_odd(self, arctic_parallel):
            image_pre_cti = np.zeros((5, 4))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            arctic_niter_5 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=5, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_5)

            image_difference_niter_5 = image_correct_cti - image_pre_cti

            arctic_niter_3 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=3, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_3)

            image_difference_niter_3 = image_correct_cti - image_pre_cti

            # noinspection PyUnresolvedReferences
            assert (abs(image_difference_niter_5) <= abs(
                image_difference_niter_3)).all()  # First four rows should all remain zero

        def test__dencrease_niter__worse_correction__odd_x_even(self, arctic_parallel):
            image_pre_cti = np.zeros((3, 6))
            image_pre_cti[2, :] += 100

            parameters = setup(include_parallel=True, p_trap_densities=(0.1,), p_trap_lifetimes=(1.0,),
                               p_well_notch_depth=0.01, p_well_fill_beta=0.8)

            image_post_cti = add_parallel_cti_to_image(image=image_pre_cti, params=parameters,
                                                       settings=arctic_parallel)

            arctic_niter_5 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=5, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_5)

            image_difference_niter_5 = image_correct_cti - image_pre_cti

            arctic_niter_3 = setup_settings(include_parallel=True, p_well_depth=84700, p_niter=3, p_express=5,
                                            p_n_levels=2000,
                                            p_charge_injection_mode=False, p_readout_offset=0)

            image_correct_cti = correct_parallel_cti_from_image(image=image_post_cti, params=parameters,
                                                                settings=arctic_niter_3)

            image_difference_niter_3 = image_correct_cti - image_pre_cti

            # noinspection PyUnresolvedReferences
            assert (abs(image_difference_niter_5) <= abs(
                image_difference_niter_3)).all()  # First four rows should all remain zero



from autofit.mapper import model_mapper as mm
from autofit.optimize import non_linear as nl

from autocti.pipeline import pipeline as pl


class DummyPhase(object):
    def __init__(self):
        self.masked_image = None
        self.previous_results = None
        self.phase_name = "dummy_phase"

    def run(self, ci_datas, cti_settings, previous_results, pool):
        self.ci_datas = ci_datas
        self.cti_settings = cti_settings
        self.previous_results = previous_results
        self.pool = pool
        return nl.Result(mm.ModelInstance(), 1)


class TestPipeline(object):
    def test_run_pipeline(self):
        phase_1 = DummyPhase()
        phase_2 = DummyPhase()
        pipeline = pl.Pipeline(phase_1, phase_2)

        pipeline.run(ci_datas=None, cti_settings=None, pool=None)

        assert len(phase_1.previous_results) == 0
        assert len(phase_2.previous_results) == 1

    def test_addition(self):
        phase_1 = DummyPhase()
        phase_2 = DummyPhase()
        phase_3 = DummyPhase()

        pipeline1 = pl.Pipeline(phase_1, phase_2)
        pipeline2 = pl.Pipeline(phase_3)

        assert (phase_1, phase_2, phase_3) == (pipeline1 + pipeline2).phases



import os

from autofit import conf
from autofit.optimize import non_linear as nl
from autofit.mapper import prior_model
from autocti.charge_injection import ci_data, ci_pattern
from autocti.model import arctic_params
from autocti.model import arctic_settings

from autocti.pipeline import phase as ph
from autocti.pipeline import pipeline as pl
from test.integration import tools

from multiprocessing import Pool

shape = (36, 36)
ci_regions = [(1, 7, 1, 30), (17, 23, 1, 30)]
normalizations = [10000.0, 84700.0]
frame_geometry = tools.CIQuadGeometryIntegration()

test_type = 'parallel_and_serial'
test_name = 'x1_species_x2_images_use_pool'

path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
output_path = path+'output/'+test_type
config_path = path+'config'
conf.instance = conf.Config(config_path=config_path, output_path=output_path)

def pipeline():

    tools.reset_paths(test_name=test_name, output_path=output_path)

    parallel_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.5)
    parallel_ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=0.2,
                                     well_fill_beta=0.8, well_fill_gamma=2.0)

    serial_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.5)
    serial_ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=0.2,
                                     well_fill_beta=0.8, well_fill_gamma=2.0)

    cti_params = arctic_params.ArcticParams(parallel_ccd=parallel_ccd, parallel_species=[parallel_species],
                                            serial_ccd=serial_ccd, serial_species=[serial_species])

    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=1, n_levels=2000,
                                                 charge_injection_mode=True, readout_offset=0)
    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=1, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(parallel=parallel_settings, serial=serial_settings)

    tools.simulate_integration_quadrant(test_name=test_name, normalizations=normalizations, cti_params=cti_params,
                                        cti_settings=cti_settings)

    patterns = ci_pattern.uniform_from_lists(normalizations=normalizations, regions=ci_regions)

    data_0 = ci_data.load_ci_data_from_fits(frame_geometry=frame_geometry, ci_pattern=patterns[0],
                                            ci_image_path=path+'/data/'+test_name+'/ci_image_0.fits',
                                            ci_noise_map_from_single_value=1.0)

    data_1 = ci_data.load_ci_data_from_fits(frame_geometry=frame_geometry, ci_pattern=patterns[1],
                                            ci_image_path=path+'/data/'+test_name+'/ci_image_1.fits',
                                            ci_noise_map_from_single_value=1.0)

    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(ci_datas=[data_0, data_1], cti_settings=cti_settings, pool=Pool(processes=2))


def make_pipeline(test_name):

    class ParallelSerialPhase(ph.ParallelSerialPhase):

        def pass_priors(self, previous_results):

            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase1 = ParallelSerialPhase(optimizer_class=nl.MultiNest,
                                 parallel_species=[prior_model.PriorModel(arctic_params.Species)],
                                 parallel_ccd=arctic_params.CCD,
                                 serial_species=[prior_model.PriorModel(arctic_params.Species)],
                                 serial_ccd=arctic_params.CCD, phase_name="{}/phase1".format(test_name))
    phase1.optimizer.n_live_points = 60
    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.sampling_efficiency = 0.2

    return pl.Pipeline(phase1)


if __name__ == "__main__":
    pipeline()
